# --------------------------------------------------------------- Imports ---------------------------------------------------------------- #

# System
from typing import Optional, List, Dict, Tuple, Union
import os, pickle, json

# Pip
import tldextract

# Local
from ...models import Cookie

# ---------------------------------------------------------------------------------------------------------------------------------------- #



# ------------------------------------------------------------ class: Cookies ------------------------------------------------------------ #

class Cookies:

    # -------------------------------------------------------- Public methods -------------------------------------------------------- #

    @classmethod
    def load(
        cls,
        url: str,
        folder_path: str
    ) -> Optional[List[Cookie]]:
        saved_cookies_path = cls.saved_cookies_path(
            url=url,
            folder_path=folder_path,
            create_folder_if_not_exists=False,
            delete_old_if_both_exists=False
        )

        if not saved_cookies_path:
            return None

        cookies = pickle.load(open(saved_cookies_path, 'rb')) if saved_cookies_path.endswith('.pkl') else json.load(open(saved_cookies_path, 'rb'))

        return [Cookie(cookie_dict) for cookie_dict in cookies]

    @classmethod
    def save(
        cls,
        url: str,
        folder_path: str,
        cookies: List[Union[Cookie, Dict[str, any]]],
        pickled: bool
    ) -> bool:
        if not cookies:
            print('Value passed for \'cookies\' is None or empty so will not save it. Use \'.delete_cookies\' to delete it.')

            return False

        cookies_path = cls.cookies_path(
            url=url,
            folder_path=folder_path,
            create_folder_if_not_exists=True,
            pickled=pickled,
            delete_old_if_both_exists=True
        )

        if os.path.exists(cookies_path):
            os.remove(cookies_path)

        _cookies = [c.dict if isinstance(c, Cookie) else c for c in cookies]

        pickle.dump(
            _cookies,
            open(cookies_path, 'wb')
        ) if pickled else json.dump(
            _cookies,
            open(cookies_path, 'w')
        )

        return True

    @classmethod
    def delete(
        cls,
        url: str,
        folder_path: str
    ) -> bool:
        did_remove_any_cookies = False

        for p in cls.possible_saved_cookies_paths(
            url=url,
            folder_path=folder_path,
            create_folder_if_not_exists=False,
            delete_old_if_both_exists=False
        ):
            if os.path.exists(p):
                did_remove_any_cookies = True
                os.remove(p)

        if not did_remove_any_cookies:
            print('No cookies found to delete for url \'{}\''.format(url))

        return did_remove_any_cookies

    @classmethod
    def has_saved_cookies(
        cls,
        url: str,
        folder_path: str
    ) -> bool:
        path = cls.saved_cookies_path(
            url=url,
            folder_path=folder_path,
            create_folder_if_not_exists=False,
            delete_old_if_both_exists=False
        )

        return path and os.path.exists(path)

    @classmethod
    def saved_cookies_path(
        cls,
        url: str,
        folder_path: str,
        create_folder_if_not_exists: bool,
        delete_old_if_both_exists: bool
    ) -> Optional[str]:
        p_pickled, p_json = cls.possible_saved_cookies_paths(
            url=url,
            folder_path=folder_path,
            create_folder_if_not_exists=create_folder_if_not_exists,
            delete_old_if_both_exists=delete_old_if_both_exists
        )

        pickled_exists = os.path.exists(p_pickled)
        json_exists = os.path.exists(p_json)

        if not delete_old_if_both_exists and pickled_exists and json_exists:
            return p_pickled if os.path.getmtime(p_pickled) > os.path.getmtime(p_json) else p_json

        return p_pickled if pickled_exists else p_json if json_exists else None

    @classmethod
    def cookies_path(
        cls,
        url: str,
        folder_path: str,
        create_folder_if_not_exists: bool,
        pickled: bool,
        delete_old_if_both_exists: bool
    ) -> str:
        return cls.possible_saved_cookies_paths(
            url=url,
            folder_path=folder_path,
            create_folder_if_not_exists=create_folder_if_not_exists,
            delete_old_if_both_exists=delete_old_if_both_exists
        )[0 if pickled else 1]

    @staticmethod
    def possible_saved_cookies_paths(
        url: str,
        folder_path: str,
        create_folder_if_not_exists: bool,
        delete_old_if_both_exists: bool
    ) -> Tuple[str, str]:
        url_comps = tldextract.extract(url)

        if create_folder_if_not_exists and not os.path.exists(folder_path):
            os.makedirs(folder_path)

        base_path = os.path.join(
            folder_path,
            '{}.{}'.format(
                url_comps.domain,
                url_comps.suffix,
            )
        )

        p_picked, p_json = '{}.pkl'.format(base_path), '{}.json'.format(base_path)

        if delete_old_if_both_exists and os.path.exists(p_picked) and os.path.exists(p_json):
            os.remove(p_json if os.path.getmtime(p_picked) > os.path.getmtime(p_json) else p_picked)

        return p_picked, p_json


# ---------------------------------------------------------------------------------------------------------------------------------------- #


from .firefox_cookies import FirefoxCookies
from .firefox_driver_wraps import FirefoxDriverWraps
from .firefox_find_funcs import FirefoxFindFuncs
from .firefox_js_functions import FirefoxJSFunctions
from .firefox_properties import FirefoxProperties
from .firefox_webelement_functions import FirefoxWebelementFunctions


# --------------------------------------------------------------- Imports ---------------------------------------------------------------- #

# System
from typing import Optional, List, Union, Dict, Callable
import json

# Pip
from selenium.webdriver.firefox.webdriver import WebDriver
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC

# ---------------------------------------------------------------------------------------------------------------------------------------- #



# -------------------------------------------------------- class: FirefoxFindFuncs ------------------------------------------------------- #

class FirefoxFindFuncs:

    # ------------------------------------------------------ Public properties ------------------------------------------------------- #

    driver                   : WebDriver
    default_find_func_timeout: float


    # -------------------------------------------------------- Public methods -------------------------------------------------------- #

    def find(
        self,
        by: By,
        key: str,
        element: Optional = None,
        timeout: Optional[int] = None
    ) -> Union[Optional[WebElement], List[WebElement]]:
        return self.__find(
            by,
            EC.presence_of_element_located,
            key,
            element=element,
            timeout=timeout
        )

    def find_by(
        self,
        type_: Optional[str] = None, #div, a, span, ...
        attributes: Optional[Dict[str, str]] = None,
        id_: Optional[str] = None,
        class_: Optional[str] = None,
        in_element: Optional[WebElement] = None,
        timeout: Optional[int] = None,
        **kwargs
    ) -> Optional[WebElement]:
        return self.find(
            By.XPATH,
            self.generate_xpath(type_=type_, attributes=attributes, id_=id_, class_=class_, for_sub_element=in_element is not None, **kwargs),
            element=in_element,
            timeout=timeout
        )

    # aliases
    bsfind = find_by
    find_ = find_by


    def find_all(
        self,
        by: By,
        key: str,
        element: Optional = None,
        timeout: Optional[int] = None
    ) -> List[WebElement]:
        return self.__find(
            by,
            EC.presence_of_all_elements_located,
            key,
            element=element,
            timeout=timeout
        )

    def find_all_by(
        self,
        type_: Optional[str] = None, #div, a, span, ...
        attributes: Optional[Dict[str, str]] = None,
        id_: Optional[str] = None,
        class_: Optional[str] = None,
        in_element: Optional[WebElement] = None,
        timeout: Optional[int] = None,
        **kwargs
    ) -> List[WebElement]:
        return self.find_all(
            By.XPATH,
            self.generate_xpath(type_=type_, attributes=attributes, id_=id_, class_=class_, for_sub_element=in_element is not None, **kwargs),
            element=in_element,
            timeout=timeout
        )

    # aliases
    bsfind_all = find_all_by
    find_all_ = find_all_by


    def reload_element(
        self,
        element,
        timeout: Optional[int] = None
    ) -> Optional[WebElement]:
        return self.find(
            By.XPATH,
            key=element.get_xpath(),
            timeout=timeout
        )

    @staticmethod
    def generate_xpath(
        type_: Optional[str] = None, #div, a, span, ...
        attributes: Optional[Dict[str, str]] = None,
        id_: Optional[str] = None,
        class_: Optional[str] = None,
        for_sub_element: bool = False, # selenium has a bug with xpath. If xpath does not start with '.' it will search in the whole doc
        **kwargs
    ) -> str:
        attributes = attributes or {}

        if class_ is not None:
            attributes['class'] = class_

        if id_ is not None:
            attributes['id'] = id_

        attributes.update({k:(v if type(v) == str else json.dumps(v)) for k, v in kwargs.items()})
        type_ = type_ or '*'
        xpath_query = ''

        for key, value in attributes.items():
            if len(xpath_query) > 0:
                xpath_query += ' and '

            xpath_query += '@' + key + '=\'' + value + '\''

        return ('.' if for_sub_element else '') + '//' + type_ + (('[' + xpath_query + ']') if len(xpath_query) > 0 else '')


    # ------------------------------------------------------- Private methods -------------------------------------------------------- #

    def __find(
        self,
        by: By,
        find_func: Callable,
        key: str,
        element: Optional = None,
        timeout: Optional[int] = None
    ) -> Union[Optional[WebElement], List[WebElement]]:
        timeout = timeout if timeout is not None else self.default_find_func_timeout

        if element is None:
            element = self.driver
        elif by == By.XPATH and not key.startswith('.'):
            # selenium has a bug with xpath. If xpath does not start with '.' it will search in the whole doc
            key = '.' + key

        try:
            es = WebDriverWait(element, timeout).until(
                find_func((by, key))
            )

            return es
        except:
            return None


# ---------------------------------------------------------------------------------------------------------------------------------------- #


# --------------------------------------------------------------- Imports ---------------------------------------------------------------- #

# System
from typing import Dict

# ---------------------------------------------------------------------------------------------------------------------------------------- #



# ------------------------------------------------------------- class: Prefs ------------------------------------------------------------- #

class Prefs:

    # ------------------------------------------------------------- Init ------------------------------------------------------------- #

    def __init__(
        self,
        d: Dict[str, any]
    ):
        self.app_normandy_api_url                                                = d.get('app.normandy.api_url')
        self.app_normandy_first_run                                              = d.get('app.normandy.first_run')
        self.app_normandy_migrations_applied                                     = d.get('app.normandy.migrationsApplied')
        self.app_shield_optoutstudies_enabled                                    = d.get('app.shield.optoutstudies.enabled')
        self.app_update_auto                                                     = d.get('app.update.auto')
        self.app_update_check_install_time                                       = d.get('app.update.checkInstallTime')
        self.app_update_disabled_for_testing                                     = d.get('app.update.disabledForTesting')
        self.app_update_enabled                                                  = d.get('app.update.enabled')
        self.app_update_last_update_time_addon_background_update_timer           = d.get('app.update.lastUpdateTime.addon-background-update-timer')
        self.app_update_last_update_time_background_update_timer                 = d.get('app.update.lastUpdateTime.background-update-timer')
        self.app_update_last_update_time_browser_cleanup_thumbnails              = d.get('app.update.lastUpdateTime.browser-cleanup-thumbnails')
        self.app_update_last_update_time_region_update_timer                     = d.get('app.update.lastUpdateTime.region-update-timer')
        self.app_update_last_update_time_search_engine_update_timer              = d.get('app.update.lastUpdateTime.search-engine-update-timer')
        self.app_update_last_update_time_services_settings_poll_changes          = d.get('app.update.lastUpdateTime.services-settings-poll-changes')
        self.app_update_last_update_time_telemetry_modules_ping                  = d.get('app.update.lastUpdateTime.telemetry_modules_ping')
        self.app_update_last_update_time_xpi_signature_verification              = d.get('app.update.lastUpdateTime.xpi-signature-verification')
        self.apz_content_response_timeout                                        = d.get('apz.content_response_timeout')
        self.browser_eula_3_accepted                                             = d.get('browser.EULA.3.accepted')
        self.browser_eula_override                                               = d.get('browser.EULA.override')
        self.browser_bookmarks_added_import_button                               = d.get('browser.bookmarks.addedImportButton')
        self.browser_bookmarks_restore_default_bookmarks                         = d.get('browser.bookmarks.restore_default_bookmarks')
        self.browser_contentblocking_category                                    = d.get('browser.contentblocking.category')
        self.browser_contentblocking_intro_count                                 = d.get('browser.contentblocking.introCount')
        self.browser_displayed_e10_snotice                                       = d.get('browser.displayedE10SNotice')
        self.browser_dom_window_dump_enabled                                     = d.get('browser.dom.window.dump.enabled')
        self.browser_download_manager_show_when_starting                         = d.get('browser.download.manager.showWhenStarting')
        self.browser_download_panel_shown                                        = d.get('browser.download.panel.shown')
        self.browser_download_viewable_internally_type_was_registered_svg        = d.get('browser.download.viewableInternally.typeWasRegistered.svg')
        self.browser_download_viewable_internally_type_was_registered_webp       = d.get('browser.download.viewableInternally.typeWasRegistered.webp')
        self.browser_download_viewable_internally_type_was_registered_xml        = d.get('browser.download.viewableInternally.typeWasRegistered.xml')
        self.browser_link_open_external                                          = d.get('browser.link.open_external')
        self.browser_link_open_newwindow                                         = d.get('browser.link.open_newwindow')
        self.browser_migration_version                                           = d.get('browser.migration.version')
        self.browser_newtab_url                                                  = d.get('browser.newtab.url')
        self.browser_newtabpage_activity_stream_impression_id                    = d.get('browser.newtabpage.activity-stream.impressionId')
        self.browser_newtabpage_enabled                                          = d.get('browser.newtabpage.enabled')
        self.browser_newtabpage_storage_version                                  = d.get('browser.newtabpage.storageVersion')
        self.browser_offline                                                     = d.get('browser.offline')
        self.browser_page_actions_persisted_actions                              = d.get('browser.pageActions.persistedActions')
        self.browser_pagethumbnails_capturing_disabled                           = d.get('browser.pagethumbnails.capturing_disabled')
        self.browser_pagethumbnails_storage_version                              = d.get('browser.pagethumbnails.storage_version')
        self.browser_reader_detected_first_article                               = d.get('browser.reader.detectedFirstArticle')
        self.browser_region_update_updated                                       = d.get('browser.region.update.updated')
        self.browser_safebrowsing_blocked_ur_is_enabled                          = d.get('browser.safebrowsing.blockedURIs.enabled')
        self.browser_safebrowsing_downloads_enabled                              = d.get('browser.safebrowsing.downloads.enabled')
        self.browser_safebrowsing_enabled                                        = d.get('browser.safebrowsing.enabled')
        self.browser_safebrowsing_malware_enabled                                = d.get('browser.safebrowsing.malware.enabled')
        self.browser_safebrowsing_phishing_enabled                               = d.get('browser.safebrowsing.phishing.enabled')
        self.browser_safebrowsing_provider_mozilla_lastupdatetime                = d.get('browser.safebrowsing.provider.mozilla.lastupdatetime')
        self.browser_safebrowsing_provider_mozilla_nextupdatetime                = d.get('browser.safebrowsing.provider.mozilla.nextupdatetime')
        self.browser_search_region                                               = d.get('browser.search.region')
        self.browser_search_update                                               = d.get('browser.search.update')
        self.browser_selfsupport_url                                             = d.get('browser.selfsupport.url')
        self.browser_sessionstore_resume_from_crash                              = d.get('browser.sessionstore.resume_from_crash')
        self.browser_shell_check_default_browser                                 = d.get('browser.shell.checkDefaultBrowser')
        self.browser_slow_startup_average_time                                   = d.get('browser.slowStartup.averageTime')
        self.browser_slow_startup_samples                                        = d.get('browser.slowStartup.samples')
        self.browser_startup_homepage                                            = d.get('browser.startup.homepage')
        self.browser_startup_homepage_override_mstone                            = d.get('browser.startup.homepage_override.mstone')
        self.browser_startup_last_cold_startup_check                             = d.get('browser.startup.lastColdStartupCheck')
        self.browser_startup_page                                                = d.get('browser.startup.page')
        self.browser_tabs_close_window_with_last_tab                             = d.get('browser.tabs.closeWindowWithLastTab')
        self.browser_tabs_disable_background_zombification                       = d.get('browser.tabs.disableBackgroundZombification')
        self.browser_tabs_remote_separate_privileged_content_process             = d.get('browser.tabs.remote.separatePrivilegedContentProcess')
        self.browser_tabs_warn_on_close                                          = d.get('browser.tabs.warnOnClose')
        self.browser_tabs_warn_on_close_other_tabs                               = d.get('browser.tabs.warnOnCloseOtherTabs')
        self.browser_tabs_warn_on_open                                           = d.get('browser.tabs.warnOnOpen')
        self.browser_toolbars_bookmarks_visibility                               = d.get('browser.toolbars.bookmarks.visibility')
        self.browser_ui_customization_state                                      = d.get('browser.uiCustomization.state')
        self.browser_uitour_enabled                                              = d.get('browser.uitour.enabled')
        self.browser_urlbar_placeholder_name                                     = d.get('browser.urlbar.placeholderName')
        self.browser_urlbar_suggest_searches                                     = d.get('browser.urlbar.suggest.searches')
        self.browser_used_on_windows10_intro_ur_l                                = d.get('browser.usedOnWindows10.introURL')
        self.browser_warn_on_quit                                                = d.get('browser.warnOnQuit')
        self.datareporting_healthreport_about_report_url                         = d.get('datareporting.healthreport.about.reportUrl')
        self.datareporting_healthreport_document_server_uri                      = d.get('datareporting.healthreport.documentServerURI')
        self.datareporting_healthreport_logging_console_enabled                  = d.get('datareporting.healthreport.logging.consoleEnabled')
        self.datareporting_healthreport_service_enabled                          = d.get('datareporting.healthreport.service.enabled')
        self.datareporting_healthreport_service_first_run                        = d.get('datareporting.healthreport.service.firstRun')
        self.datareporting_healthreport_upload_enabled                           = d.get('datareporting.healthreport.uploadEnabled')
        self.datareporting_policy_data_submission_enabled                        = d.get('datareporting.policy.dataSubmissionEnabled')
        self.datareporting_policy_data_submission_policy_accepted                = d.get('datareporting.policy.dataSubmissionPolicyAccepted')
        self.datareporting_policy_data_submission_policy_bypass_notification     = d.get('datareporting.policy.dataSubmissionPolicyBypassNotification')
        self.devtools_console_stdout_chrome                                      = d.get('devtools.console.stdout.chrome')
        self.devtools_errorconsole_enabled                                       = d.get('devtools.errorconsole.enabled')
        self.distribution_ini_file_exists_appversion                             = d.get('distribution.iniFile.exists.appversion')
        self.distribution_ini_file_exists_value                                  = d.get('distribution.iniFile.exists.value')
        self.doh_rollout_balrog_migration_done                                   = d.get('doh-rollout.balrog-migration-done')
        self.doh_rollout_done_first_run                                          = d.get('doh-rollout.doneFirstRun')
        self.dom_disable_beforeunload                                            = d.get('dom.disable_beforeunload')
        self.dom_disable_open_during_load                                        = d.get('dom.disable_open_during_load')
        self.dom_file_create_in_child                                            = d.get('dom.file.createInChild')
        self.dom_ipc_report_process_hangs                                        = d.get('dom.ipc.reportProcessHangs')
        self.dom_max_chrome_script_run_time                                      = d.get('dom.max_chrome_script_run_time')
        self.dom_max_script_run_time                                             = d.get('dom.max_script_run_time')
        self.dom_push_connection_enabled                                         = d.get('dom.push.connection.enabled')
        self.dom_report_all_js_exceptions                                        = d.get('dom.report_all_js_exceptions')
        self.dom_webdriver_enabled                                               = d.get('dom.webdriver.enabled')
        self.extensions_active_theme_id                                          = d.get('extensions.activeThemeID')
        self.extensions_auto_disable_scopes                                      = d.get('extensions.autoDisableScopes')
        self.extensions_blocklist_enabled                                        = d.get('extensions.blocklist.enabled')
        self.extensions_blocklist_ping_count_version                             = d.get('extensions.blocklist.pingCountVersion')
        self.extensions_check_compatibility_nightly                              = d.get('extensions.checkCompatibility.nightly')
        self.extensions_database_schema                                          = d.get('extensions.databaseSchema')
        self.extensions_enabled_scopes                                           = d.get('extensions.enabledScopes')
        self.extensions_get_addons_cache_enabled                                 = d.get('extensions.getAddons.cache.enabled')
        self.extensions_get_addons_database_schema                               = d.get('extensions.getAddons.databaseSchema')
        self.extensions_get_addons_discovery_api_url                             = d.get('extensions.getAddons.discovery.api_url')
        self.extensions_incognito_migrated                                       = d.get('extensions.incognito.migrated')
        self.extensions_install_distro_addons                                    = d.get('extensions.installDistroAddons')
        self.extensions_last_app_build_id                                        = d.get('extensions.lastAppBuildId')
        self.extensions_last_app_version                                         = d.get('extensions.lastAppVersion')
        self.extensions_last_platform_version                                    = d.get('extensions.lastPlatformVersion')
        self.extensions_pending_operations                                       = d.get('extensions.pendingOperations')
        self.extensions_show_mismatch_ui                                         = d.get('extensions.showMismatchUI')
        self.extensions_system_addon_set                                         = d.get('extensions.systemAddonSet')
        self.extensions_update_enabled                                           = d.get('extensions.update.enabled')
        self.extensions_update_notify_user                                       = d.get('extensions.update.notifyUser')
        self.extensions_webcompat_enable_picture_in_picture_overrides            = d.get('extensions.webcompat.enable_picture_in_picture_overrides')
        self.extensions_webcompat_enable_shims                                   = d.get('extensions.webcompat.enable_shims')
        self.extensions_webcompat_perform_injections                             = d.get('extensions.webcompat.perform_injections')
        self.extensions_webcompat_perform_ua_overrides                           = d.get('extensions.webcompat.perform_ua_overrides')
        self.extensions_webextensions_uuids                                      = d.get('extensions.webextensions.uuids')
        self.focusmanager_testmode                                               = d.get('focusmanager.testmode')
        self.general_useragent_updates_enabled                                   = d.get('general.useragent.updates.enabled')
        self.general_warn_on_about_config                                        = d.get('general.warnOnAboutConfig')
        self.geo_provider_testing                                                = d.get('geo.provider.testing')
        self.geo_wifi_scan                                                       = d.get('geo.wifi.scan')
        self.hangmonitor_timeout                                                 = d.get('hangmonitor.timeout')
        self.intl_accept_languages                                               = d.get('intl.accept_languages')
        self.marionatte                                                          = d.get('marionatte')
        self.marionette_content_listener                                         = d.get('marionette.contentListener')
        self.marionette_enabled                                                  = d.get('marionette.enabled')
        self.marionette_port                                                     = d.get('marionette.port')
        self.media_gmp_manager_build_id                                          = d.get('media.gmp-manager.buildID')
        self.media_gmp_manager_last_check                                        = d.get('media.gmp-manager.lastCheck')
        self.media_gmp_manager_update_enabled                                    = d.get('media.gmp-manager.updateEnabled')
        self.media_gmp_storage_version_observed                                  = d.get('media.gmp.storage.version.observed')
        self.media_peerconnection_enabled                                        = d.get('media.peerconnection.enabled')
        self.network_captive_portal_service_enabled                              = d.get('network.captive-portal-service.enabled')
        self.network_http_phishy_userpass_length                                 = d.get('network.http.phishy-userpass-length')
        self.network_manage_offline_status                                       = d.get('network.manage-offline-status')
        self.network_sntp_pools                                                  = d.get('network.sntp.pools')
        self.network_stricttransportsecurity_preloadlist                         = d.get('network.stricttransportsecurity.preloadlist')
        self.network_trr_blocklist_cleanup_done                                  = d.get('network.trr.blocklist_cleanup_done')
        self.offline_apps_allow_by_default                                       = d.get('offline-apps.allow_by_default')
        self.pdfjs_enabled_cache_state                                           = d.get('pdfjs.enabledCache.state')
        self.pdfjs_migration_version                                             = d.get('pdfjs.migrationVersion')
        self.plugin_state_flash                                                  = d.get('plugin.state.flash')
        self.privacy_sanitize_pending                                            = d.get('privacy.sanitize.pending')
        self.prompts_tab_modal_enabled                                           = d.get('prompts.tab_modal.enabled')
        self.security_cert_pinning_enforcement_level                             = d.get('security.cert_pinning.enforcement_level')
        self.security_certerrors_mitm_priming_enabled                            = d.get('security.certerrors.mitm.priming.enabled')
        self.security_csp_enable                                                 = d.get('security.csp.enable')
        self.security_fileuri_origin_policy                                      = d.get('security.fileuri.origin_policy')
        self.security_fileuri_strict_origin_policy                               = d.get('security.fileuri.strict_origin_policy')
        self.security_notification_enable_delay                                  = d.get('security.notification_enable_delay')
        self.security_remote_settings_crlite_filters_checked                     = d.get('security.remote_settings.crlite_filters.checked')
        self.security_remote_settings_intermediates_checked                      = d.get('security.remote_settings.intermediates.checked')
        self.security_sandbox_content_temp_dir_suffix                            = d.get('security.sandbox.content.tempDirSuffix')
        self.security_sandbox_plugin_temp_dir_suffix                             = d.get('security.sandbox.plugin.tempDirSuffix')
        self.services_blocklist_pinning_checked                                  = d.get('services.blocklist.pinning.checked')
        self.services_settings_clock_skew_seconds                                = d.get('services.settings.clock_skew_seconds')
        self.services_settings_last_etag                                         = d.get('services.settings.last_etag')
        self.services_settings_last_update_seconds                               = d.get('services.settings.last_update_seconds')
        self.services_settings_main_anti_tracking_url_decoration_last_check      = d.get('services.settings.main.anti-tracking-url-decoration.last_check')
        self.services_settings_main_cfr_fxa_last_check                           = d.get('services.settings.main.cfr-fxa.last_check')
        self.services_settings_main_cfr_last_check                               = d.get('services.settings.main.cfr.last_check')
        self.services_settings_main_fxmonitor_breaches_last_check                = d.get('services.settings.main.fxmonitor-breaches.last_check')
        self.services_settings_main_hijack_blocklists_last_check                 = d.get('services.settings.main.hijack-blocklists.last_check')
        self.services_settings_main_language_dictionaries_last_check             = d.get('services.settings.main.language-dictionaries.last_check')
        self.services_settings_main_message_groups_last_check                    = d.get('services.settings.main.message-groups.last_check')
        self.services_settings_main_normandy_recipes_capabilities_last_check     = d.get('services.settings.main.normandy-recipes-capabilities.last_check')
        self.services_settings_main_password_recipes_last_check                  = d.get('services.settings.main.password-recipes.last_check')
        self.services_settings_main_pioneer_study_addons_v1_last_check           = d.get('services.settings.main.pioneer-study-addons-v1.last_check')
        self.services_settings_main_public_suffix_list_last_check                = d.get('services.settings.main.public-suffix-list.last_check')
        self.services_settings_main_search_config_last_check                     = d.get('services.settings.main.search-config.last_check')
        self.services_settings_main_search_default_override_allowlist_last_check = d.get('services.settings.main.search-default-override-allowlist.last_check')
        self.services_settings_main_search_telemetry_last_check                  = d.get('services.settings.main.search-telemetry.last_check')
        self.services_settings_main_sites_classification_last_check              = d.get('services.settings.main.sites-classification.last_check')
        self.services_settings_main_top_sites_last_check                         = d.get('services.settings.main.top-sites.last_check')
        self.services_settings_main_url_classifier_skip_urls_last_check          = d.get('services.settings.main.url-classifier-skip-urls.last_check')
        self.services_settings_main_whats_new_panel_last_check                   = d.get('services.settings.main.whats-new-panel.last_check')
        self.services_settings_security_onecrl_checked                           = d.get('services.settings.security.onecrl.checked')
        self.services_settings_server                                            = d.get('services.settings.server')
        self.signon_autofill_forms                                               = d.get('signon.autofillForms')
        self.signon_remember_signons                                             = d.get('signon.rememberSignons')
        self.startup_homepage_welcome_url                                        = d.get('startup.homepage_welcome_url')
        self.startup_homepage_welcome_url_additional                             = d.get('startup.homepage_welcome_url.additional')
        self.toolkit_networkmanager_disable                                      = d.get('toolkit.networkmanager.disable')
        self.toolkit_startup_last_success                                        = d.get('toolkit.startup.last_success')
        self.toolkit_startup_max_resumed_crashes                                 = d.get('toolkit.startup.max_resumed_crashes')
        self.toolkit_telemetry_cached_client_id                                  = d.get('toolkit.telemetry.cachedClientID')
        self.toolkit_telemetry_pioneer_new_studies_available                     = d.get('toolkit.telemetry.pioneer-new-studies-available')
        self.toolkit_telemetry_previous_build_id                                 = d.get('toolkit.telemetry.previousBuildID')
        self.toolkit_telemetry_prompted                                          = d.get('toolkit.telemetry.prompted')
        self.toolkit_telemetry_rejected                                          = d.get('toolkit.telemetry.rejected')
        self.toolkit_telemetry_reportingpolicy_first_run                         = d.get('toolkit.telemetry.reportingpolicy.firstRun')
        self.toolkit_telemetry_server                                            = d.get('toolkit.telemetry.server')
        self.use_automation_extension                                            = d.get('useAutomationExtension')
        self.webdriver_accept_untrusted_certs                                    = d.get('webdriver_accept_untrusted_certs')
        self.webdriver_assume_untrusted_issuer                                   = d.get('webdriver_assume_untrusted_issuer')
        self.webdriver_enable_native_events                                      = d.get('webdriver_enable_native_events')
        self.xpinstall_signatures_required                                       = d.get('xpinstall.signatures.required')
        self.xpinstall_whitelist_required                                        = d.get('xpinstall.whitelist.required')

        self.extensions_webextensions_extension_storage_idb_migrated_foxyproxy_eric_h_jung   = d.get('extensions.webextensions.ExtensionStorageIDB.migrated.foxyproxy@eric.h.jung')
        self.extensions_webextensions_extension_storage_idb_migrated_screenshots_mozilla_org = d.get('extensions.webextensions.ExtensionStorageIDB.migrated.screenshots@mozilla.org')

# ---------------------------------------------------------------------------------------------------------------------------------------- #


# --------------------------------------------------------------- Imports ---------------------------------------------------------------- #

# System
from typing import Optional, Tuple, Union, List
import random, time, copy

# Pip
from noraise import noraise

from selenium.webdriver.firefox.webdriver import WebDriver
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.common.action_chains import ActionChains

# Local
from .firefox_js_functions import FirefoxJSFunctions

# ---------------------------------------------------------------------------------------------------------------------------------------- #



# --------------------------------------------------- class: FirefoxWebelementFunctions -------------------------------------------------- #

class FirefoxWebelementFunctions(FirefoxJSFunctions):

    # ------------------------------------------------------ Public properties ------------------------------------------------------- #

    driver: WebDriver


    # -------------------------------------------------------- Public methods -------------------------------------------------------- #

    @noraise()
    def get_attribute(self, element, key: str) -> Optional[str]:
        return element.get_attribute(key)

    @noraise(default_return_value=False)
    def send_keys_delay_random(
        self,
        element: WebElement,
        keys: Union[str, int, float, List[str], List[int], List[float]],
        min_delay: float = 0.025,
        max_delay: float = 0.25
    ) -> bool:
        _keys = copy.deepcopy(keys)

        if type(_keys) in [int, float]:
            _keys = str(_keys)

        for key in _keys:
            if type(key) in [int, float]:
                key = str(key)

            element.send_keys(key)
            time.sleep(random.uniform(min_delay,max_delay))

        return True

    @noraise(default_return_value=False)
    def scroll_to_element(self, element, header_element=None) -> bool:
        header_h = 0

        if header_element is not None:
            _, _, _, header_h, _, _ = self.get_element_coordinates(header_element)

        _, element_y, _, _, _, _ = self.get_element_coordinates(element)

        return self.scroll_to(element_y-header_h)

    def scroll_to_bottom(self, steps_pixels: int = 1000) -> None:
        MAX_TRIES = 25
        SCROLL_PAUSE_TIME = 0.5
        current_tries = 1
        last_height = 0

        while True:
            self.scroll(steps_pixels)
            time.sleep(SCROLL_PAUSE_TIME)
            current_height = self.current_page_offset_y()

            if last_height == current_height:
                current_tries += 1

                if current_tries == MAX_TRIES:
                    break
                else:
                    continue
            else:
                current_tries = 1

            last_height = current_height


    @noraise(default_return_value=False)
    def move_to_element(
        self,
        element: Optional[WebElement],
        offset: Optional[Tuple[int, int]] = None,
        click: bool = False
    ) -> bool:
        if not element:
            print('move_to_element: None element passed')

            return False

        ac = ActionChains(self.driver)
        ac.move_to_element_with_offset(element, offset[0], offset[1]) if offset else ac.move_to_element(element)

        if click:
            ac.click()

        ac.perform()

        return True

    @noraise(default_return_value=False)
    def switch_to_frame(self, iframe: Optional[WebElement]) -> bool:
        if not iframe:
            print('switch_to_frame: None frame passed')

            return False

        self.driver.switch_to.frame(iframe)

        return True

    # returns x, y, w, h, max_x, max_y
    def get_element_coordinates(self, element: WebElement) -> Tuple[int, int, int, int, int, int]:
        location = element.location
        size = element.size

        x = location['x']
        y = location['y']
        w = size['width']
        h = size['height']

        return x, y, w, h, x+w, y+h


# ---------------------------------------------------------------------------------------------------------------------------------------- #


# --------------------------------------------------------------- Imports ---------------------------------------------------------------- #

# System
from typing import Optional, List, Tuple, Callable, Dict
import os

# Pip
from noraise import noraise

from selenium.webdriver.firefox.webdriver import WebDriver as FirefoxWebdriver

# Local
# from ..firefox import Firefox
from ..addon_install_settings import BaseAddonInstallSettings

# ---------------------------------------------------------------------------------------------------------------------------------------- #



# ---------------------------------------------------------- class: AddonManager --------------------------------------------------------- #

class AddonManager:

    # -------------------------------------------------------- Public methods -------------------------------------------------------- #

    @classmethod
    def addons_from_folder(
        cls,
        folder_path: str
    ) -> List[BaseAddonInstallSettings]:
        for (dirpath, _, filenames) in os.walk(folder_path):
            return [
                BaseAddonInstallSettings(os.path.join(dirpath, f))
                for f in filenames if f.split('.')[-1] in cls.__SUPPORTED_ADDON_FILE_EXTENSIONS
            ]

    @classmethod
    def install_addons(
        cls,
        firefox,#: Firefox,
        addons_settings: List[BaseAddonInstallSettings],
        temporary: Optional[bool] = False
    ) -> List[Optional[str]]:
        return [
            cls.install_addon(firefox=firefox, addon_settings=addon_settings, temporary=temporary)
            for addon_settings in addons_settings
        ]

    @classmethod
    def install_addon(
        cls,
        firefox,#: Firefox,
        addon_settings: BaseAddonInstallSettings,
        temporary: Optional[bool] = False
    ) -> Optional[str]:
        if addon_settings.path.split('.')[-1] not in cls.__SUPPORTED_ADDON_FILE_EXTENSIONS:
            print('Will not install \'{}\'. Reason: not one of the needed formats: {}'.format(
                addon_settings.path,
                cls.__SUPPORTED_ADDON_FILE_EXTENSIONS
            ))

            return None

        org_handle = firefox.driver.current_window_handle
        addon_id = firefox.driver.install_addon(addon_settings.path, temporary=temporary)

        if not addon_id:
            print('Could not install addon with path \'{}\''.format(addon_settings.path))

        print('Did install addon with id \'{}\' and path \'{}\''.format(addon_id, addon_settings.path))

        internal_addon_id = cls.__get_addon_internal_id(firefox, addon_id)
        org_current_url = firefox.current_url

        def reset(org_handle, org_current_url):
            if len(firefox.window_handles) > 1:
                for handle in firefox.window_handles:
                    firefox.switch_to_window(handle)

                    if internal_addon_id in firefox.current_url:
                        firefox.close()

                        return reset(org_handle, org_current_url)

            if org_handle not in firefox.window_handles:
                org_handle = firefox.window_handles[0]

            firefox.switch_to_window(org_handle)
            firefox.get(org_current_url)

        reset(org_handle, org_current_url)

        addon_settings.post_install_action(
            firefox,
            addon_id,
            internal_addon_id,
            cls.__ADDON_URL.format(internal_addon_id)
        )

        reset(org_handle, org_current_url)

        return addon_id


    # ------------------------------------------------------ Private properties ------------------------------------------------------ #

    __SUPPORTED_ADDON_FILE_EXTENSIONS = ['xpi', 'zip']
    __ADDON_URL                       = 'moz-extension://{}'


    # -------------------------------------------------------- Private methods ------------------------------------------------------- #

    # @noraise()
    @staticmethod
    def __get_addon_internal_id(
        firefox,#: Firefox,
        addon_public_id: str
    ) -> Optional[str]:
        import time
        start_time = time.time()

        while not firefox.prefs.extensions_webextensions_uuids.get(addon_public_id) and time.time() - start_time < 5:
            time.sleep(0.05)

        return firefox.prefs.extensions_webextensions_uuids.get(addon_public_id)


# ---------------------------------------------------------------------------------------------------------------------------------------- #


# --------------------------------------------------------------- Imports ---------------------------------------------------------------- #

# System
from typing import Optional, Union, List, Tuple
import os, shutil

# Pip
from noraise import noraise

from selenium.webdriver import Firefox as FirefoxWebDriver
from selenium.webdriver.firefox.firefox_binary import FirefoxBinary

# Local
from .__firefox_core import FirefoxCookies, FirefoxDriverWraps, FirefoxFindFuncs, FirefoxProperties, FirefoxWebelementFunctions
from .models import Proxy, Capabilities, Prefs

from .addon_install_settings import BaseAddonInstallSettings, FoxyProxyAddonSettings
from .__resources import Constants
from .__utils import AddonManager, Utils

# ---------------------------------------------------------------------------------------------------------------------------------------- #



# ------------------------------------------------------------ class: Firefox ------------------------------------------------------------ #

class Firefox(
    FirefoxCookies,
    FirefoxDriverWraps,
    FirefoxFindFuncs,
    FirefoxProperties,
    FirefoxWebelementFunctions
):

    # ------------------------------------------------------------- Init ------------------------------------------------------------- #

    def __init__(
        self,

        # cookies
        cookies_folder_path: Optional[str] = None,
        cookies_id: Optional[str] = None,
        pickle_cookies: bool = False,

        # proxy
        proxy: Optional[Union[Proxy, str]] = None,
        # proxy - legacy (kept for convenience)
        host: Optional[str] = None,
        port: Optional[int] = None,

        # addons
        addons_folder_path: Optional[str] = None,
        addon_settings: Optional[List[BaseAddonInstallSettings]] = None,
        # addons - legacy (kept for convenience)
        extensions_folder_path: Optional[str] = None,

        # other paths
        geckodriver_path: Optional[str] = None,
        firefox_binary_path: Optional[str] = None,
        profile_path: Optional[str] = None,

        # profile settings
        private: bool = False,
        full_screen: bool = True,
        language: str = 'en-us',
        user_agent: Optional[str] = None,
        disable_images: bool = False,

        # firefox option settings
        screen_size: Optional[Tuple[int, int]] = None, # (width, height)
        headless: bool = False,
        mute_audio: bool = False,
        home_page_url: Optional[str] = None,

        # selenium-wire support
        webdriver_class: Optional = None,

        # find function
        default_find_func_timeout: int = 2.5
    ):
        '''EITHER PROVIDE 'cookies_id' OR  'cookies_folder_path'.
           IF 'cookies_folder_path' is None, 'cokies_id', will be used to calculate 'cookies_folder_path'
           IF 'cokies_id' is None, it will become 'test'

           webdriver_class: override class used to create webdriver (for example: seleniumwire.webdriver.Firefox), Defaults to: 'selenium.webdriver.Firefox'
        '''

        self.default_find_func_timeout = default_find_func_timeout
        self.pickle_cookies = pickle_cookies
        self.source_profile_path = profile_path

        self.cookies_folder_path = Utils.cookies_folder_path(
            cookies_folder_path=cookies_folder_path,
            cookies_id=cookies_id
        )
        os.makedirs(self.cookies_folder_path, exist_ok=True)

        self._user_agent = Utils.user_agent(
            user_agent=user_agent,
            file_path=os.path.join(self.cookies_folder_path, Constants.USER_AGENT_FILE_NAME)
        )

        self._proxy = Utils.proxy(
            proxy=proxy,
            host=host,
            port=port
        )

        self.driver = (webdriver_class or FirefoxWebDriver)(
            firefox_profile=Utils.profile(
                user_agent=self._user_agent,
                language=language,
                private=private,
                disable_images=disable_images,
                proxy=self._proxy if self._proxy and not self._proxy.needs_auth else None,
                path=profile_path
            ),
            firefox_options=Utils.options(
                screen_size=screen_size,
                headless=headless,
                mute_audio=mute_audio,
                home_page_url=home_page_url
            ),
            firefox_binary=FirefoxBinary(
                firefox_path=firefox_binary_path
            ) if firefox_binary_path and os.path.exists(firefox_binary_path) else None,
            **Utils.webdriver_init_path_kwarg(geckodriver_path)
        )

        if full_screen:
            self.driver.fullscreen_window()

        lib_addon_settings = []

        if self._proxy and self._proxy.needs_auth:
            lib_addon_settings.append(FoxyProxyAddonSettings())

        AddonManager.install_addons(
            firefox=self,
            addons_settings=Utils.addon_settings(
                addons_folder_path=addons_folder_path or extensions_folder_path,
                user_addon_settings=addon_settings,
                lib_addon_settings=lib_addon_settings
            ),
            temporary=False
        )


    # -------------------------------------------------------- Public methods -------------------------------------------------------- #

    @noraise(default_return_value=False)
    def backup_profile(
        self,
        target_profile_path: Optional[str] = None
    ) -> bool:
        target_profile_path = target_profile_path or self.source_profile_path

        shutil.rmtree(target_profile_path)
        shutil.copytree(self.temp_profile_folder_path, target_profile_path)

        return os.path.exists(target_profile_path)


    # --------------------------------------------------------- Destructor ----------------------------------------------------------- #

    @noraise()
    def __del__(self):
        if os.path.exists(self.driver.profile.path):
            self.quit()


# ---------------------------------------------------------------------------------------------------------------------------------------- #


# --------------------------------------------------------------- Imports ---------------------------------------------------------------- #

# System
from typing import Optional
import inspect, json

# Pip
from noraise import noraise

from selenium.webdriver.firefox.webdriver import WebDriver
from selenium.webdriver.remote.webelement import WebElement

# ---------------------------------------------------------------------------------------------------------------------------------------- #



# ------------------------------------------------------- class: FirefoxJSFunctions ------------------------------------------------------ #

class FirefoxJSFunctions:

    # ------------------------------------------------------ Public properties ------------------------------------------------------- #

    driver: WebDriver


    # -------------------------------------------------------- Public methods -------------------------------------------------------- #

    def js_get_attribute(
        self,
        element: WebElement,
        key: str
    ) -> Optional:
        attrs = self.js_get_attributes(element)

        return attrs.get(key) if attrs else None

    @noraise()
    def js_get_attributes(
        self,
        element: WebElement
    ) -> Optional:
        return json.loads(
            self.execute_script_on_element(
                'var items = {}; for (index = 0; index < arguments[0].attributes.length; ++index) { items[arguments[0].attributes[index].name] = arguments[0].attributes[index].value }; return JSON.stringify(items);',
                element
            )
        )
    # alias - kept for convenience
    get_attributes = js_get_attributes

    def js_click(
        self,
        element: WebElement
    ) -> bool:
        return self.execute_script_on_element('arguments[0].click();', element)

    def js_scroll_into_view(
        self,
        element: WebElement
    ) -> bool:
        return self.execute_script_on_element('arguments[0].scrollIntoView();', element)


    def js_get_user_agent(self) -> str:
        return self.execute_script("return navigator.userAgent;")


    def current_page_offset_y(self) -> float:
        return self.execute_script('return window.pageYOffset;')

    def scroll(
        self,
        amount: int
    ) -> None:
        self.scroll_to(self.current_page_offset_y()+amount)

    def scroll_to(
        self,
        position: int
    ) -> None:
        self.execute_script('window.scrollTo(0,{});'.format(position))


    def open_new_tab(
        self,
        url: Optional[str] = None
    ) -> bool:
        if url is None:
            url = ''

        if not self.execute_script('window.open("{}","_blank");'.format(url)):
            return False

        self.driver.switch_to.window(self.driver.window_handles[-1])

        return True


    def execute_script_on_element(
        self,
        script: str,
        element: WebElement
    ) -> bool:
        caller_name = inspect.stack()[2][3]

        if not element:
            print('{}: passed element is None'.format(caller_name))

            return False

        return self.execute_script(script, element)

    @noraise(default_return_value=False)
    def execute_script(
        self,
        script: str,
        element: Optional[WebElement] = None
    ) -> Optional[any]:
        res = self.driver.execute_script(script, element) if element else self.driver.execute_script(script)

        return res or True


# ---------------------------------------------------------------------------------------------------------------------------------------- #


# --------------------------------------------------------------- Imports ---------------------------------------------------------------- #

# System
from typing import Dict

# ---------------------------------------------------------------------------------------------------------------------------------------- #



# ---------------------------------------------------------- class: Capabilities --------------------------------------------------------- #

class Capabilities:

    # ------------------------------------------------------------- Init ------------------------------------------------------------- #

    def __init__(
        self,
        d: Dict[str, any]
    ):
        self.accept_insecure_certs                     = d.get('acceptInsecureCerts')
        self.browser_name                              = d.get('browserName')
        self.browser_version                           = d.get('browserVersion')
        self.moz_accessibility_checks                  = d.get('moz:accessibilityChecks')
        self.moz_build_id                              = d.get('moz:buildID')
        self.moz_geckodriver_version                   = d.get('moz:geckodriverVersion')
        self.moz_headless                              = d.get('moz:headless')
        self.moz_process_id                            = d.get('moz:processID')
        self.moz_profile                               = d.get('moz:profile')
        self.moz_shutdown_timeout                      = d.get('moz:shutdownTimeout')
        self.moz_use_non_spec_compliant_pointer_origin = d.get('moz:useNonSpecCompliantPointerOrigin')
        self.moz_webdriver_click                       = d.get('moz:webdriverClick')
        self.page_load_strategy                        = d.get('pageLoadStrategy')
        self.platform_name                             = d.get('platformName')
        self.platform_version                          = d.get('platformVersion')
        self.rotatable                                 = d.get('rotatable')
        self.set_window_rect                           = d.get('setWindowRect')
        self.strict_file_interactability               = d.get('strictFileInteractability')
        self.timeouts                                  = d.get('timeouts')
        self.unhandled_prompt_behavior                 = d.get('unhandledPromptBehavior')


# ---------------------------------------------------------------------------------------------------------------------------------------- #


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="selenium_firefox",
    version="1.0.16",
    author="Kovács Kristóf-Attila",
    description="selenium_firefox",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/kkristof200/selenium_firefox",
    packages=setuptools.find_packages(),
    install_requires=[
        'geckodriver-autoinstaller>=0.1.0',
        'noraise>=0.0.9',
        'selenium>=3.141.0',
        'tldextract>=3.1.0'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    include_package_data=True
)


# --------------------------------------------------------------- Imports ---------------------------------------------------------------- #

# System
from  typing import Optional, Dict, Tuple, Union, List
import json, os, tempfile, platform, subprocess

# Pip
from selenium.webdriver import FirefoxProfile
from selenium.webdriver.firefox.options import Options as FirefoxOptions
import geckodriver_autoinstaller

# Local
from ..models import Proxy
from ..__resources.constants import Constants
from ..addon_install_settings import BuiltinAddonInstallSettings, BaseAddonInstallSettings
from .addon_manager import AddonManager

# ---------------------------------------------------------------------------------------------------------------------------------------- #



# ------------------------------------------------------------- class: Utils ------------------------------------------------------------- #

class Utils:

    # -------------------------------------------------------- Public methods -------------------------------------------------------- #

    @classmethod
    def webdriver_init_path_kwarg(
        cls,
        user_passed_geckodriver_path: Optional[str] = None
    ) -> Dict[str, str]:
        firefox_kwargs = {}

        if user_passed_geckodriver_path and os.path.exists(user_passed_geckodriver_path):
            firefox_kwargs[Constants.WEBDRIVER_INIT_EXECUTABLE_PATH_KEYWORD] = user_passed_geckodriver_path
        elif not cls.is_geckodriver_installed():
            firefox_kwargs[Constants.WEBDRIVER_INIT_EXECUTABLE_PATH_KEYWORD] = geckodriver_autoinstaller.install()

        return firefox_kwargs

    @staticmethod
    def is_geckodriver_installed() -> bool:
        return subprocess.getstatusoutput('geckodriver -V')[0] == 0

    @staticmethod
    def cookies_folder_path(
        cookies_folder_path: Optional[str] = None,
        cookies_id: Optional[str] = None
    ) -> str:
        return cookies_folder_path or os.path.join(
            '/tmp' if platform.system() == 'Darwin' else tempfile.gettempdir(),
            Constants.GENERAL_COOKIES_FOLDER_NAME,
            cookies_id or Constants.DEFAULT_COOKIES_ID
        )

    @staticmethod
    def user_agent(
        user_agent: Optional[str] = None,
        file_path: Optional[str] = None
    ) -> Optional[str]:
        if file_path and os.path.exists(file_path):
            with open(file_path, 'r') as f:
                return f.read().strip()
        elif user_agent:
            user_agent = user_agent.strip()

            if file_path:
                with open(file_path, 'w') as f:
                    f.write(user_agent)

            return user_agent

    @staticmethod
    def proxy(
        proxy: Optional[Union[Proxy, str]] = None,

        # proxy - legacy (kept for convenience)
        host: Optional[str] = None,
        port: Optional[int] = None,
    ) -> Optional[Proxy]:
        if not proxy:
            if not host and not port:
                return None

            proxy = Proxy(host=host, port=port)

        return proxy if isinstance(proxy, Proxy) else Proxy.from_str(proxy)

    @staticmethod
    def profile(
        user_agent: Optional[str] = None,
        language: str = 'en-us',
        private: bool = False,
        disable_images: bool = False,
        proxy: Optional[Proxy] = None,
        path: Optional[str] = None
    ) -> FirefoxProfile:
        profile = FirefoxProfile(path if path and os.path.exists(path) else None)

        if user_agent:
            profile.set_preference('general.useragent.override', user_agent)

        if language:
            profile.set_preference('intl.accept_languages', language)

        if private:
            profile.set_preference('browser.privatebrowsing.autostart', True)

        if disable_images:
            profile.set_preference('permissions.default.image', 2)
            profile.set_preference('dom.ipc.plugins.enabled.libflashplayer.so', False)

        if proxy:
            profile.set_preference('network.proxy.type', 1)
            profile.set_preference('network.proxy.http', proxy.host)
            profile.set_preference('network.proxy.http_port', proxy.port)
            profile.set_preference('network.proxy.ssl', proxy.host)
            profile.set_preference('network.proxy.ssl_port', proxy.port)
            profile.set_preference('network.proxy.ftp', proxy.host)
            profile.set_preference('network.proxy.ftp_port', proxy.port)
            profile.set_preference('network.proxy.socks', proxy.host)
            profile.set_preference('network.proxy.socks_port', proxy.port)
            profile.set_preference('network.proxy.socks_version', 5)
            profile.set_preference('signon.autologin.proxy', True)

        profile.set_preference('marionatte', False)
        profile.set_preference('dom.webdriver.enabled', False)
        profile.set_preference('media.peerconnection.enabled', False)
        profile.set_preference('useAutomationExtension', False)
        profile.set_preference('general.warnOnAboutConfig', False)
        profile.update_preferences()

        return profile

    @staticmethod
    def options(
        screen_size: Optional[Tuple[int, int]] = None, # (width, height)
        headless: bool = False,
        mute_audio: bool = False,
        home_page_url: Optional[str] = None
    ) -> FirefoxOptions:
        options = FirefoxOptions()
        options.add_argument('--homepage \"{}\"'.format(home_page_url or 'data:,'))

        if screen_size is not None:
            options.add_argument('--width={}'.format(screen_size[0]))
            options.add_argument('--height={}'.format(screen_size[1]))

        if headless:
            options.add_argument('--headless')

        if mute_audio:
            options.add_argument('--mute-audio')

        return options

    @staticmethod
    def addon_settings(
        addons_folder_path: Optional[str] = None,
        user_addon_settings: Optional[List[BaseAddonInstallSettings]] = None,
        lib_addon_settings: Optional[List[BuiltinAddonInstallSettings]] = None
    ) -> List[BaseAddonInstallSettings]:
        addon_settings = lib_addon_settings or []

        if user_addon_settings:
            addon_settings.extend(user_addon_settings)

        if addons_folder_path and os.path.exists(addons_folder_path):
            addon_settings.extend(AddonManager.addons_from_folder(addons_folder_path))

        return addon_settings

    @staticmethod
    def parse_prefs_js(prefs_str: str) -> Dict[str, any]:
        prefs = {}

        for line in prefs_str.split('\n'):
            try:
                key = line.split('user_pref("')[1].split('"')[0]
                value = line.split('", ')[1].split(');')[0]

                try:
                    value = json.loads(value.replace('\\"', '"').strip('"'))
                except:
                    value = json.loads(value)

                prefs[key] = value
            except:
                pass

        return prefs


# ---------------------------------------------------------------------------------------------------------------------------------------- #


from .decentraleyes_addon_settings import DecentraleyesAddonSettings
from .foxy_proxy_addon_settings import FoxyProxyAddonSettings
from .https_everywhere_addon_settings import HttpsEverywhereAddonSettings
from .noscript_addon_settings import NoscriptAddonSettings
from .privacy_badger_addon_settings import PrivacyBadgerAddonSettings
from .ublock_origin_addon_settings import UBlockOriginAddonSettings

from .builtin_addon_install_settings import BuiltinAddonInstallSettings
from .builtin_addon_install_settings.base_addon_install_settings import BaseAddonInstallSettings


# --------------------------------------------------------------- Imports ---------------------------------------------------------------- #

# System
from typing import Optional, List, Dict, Tuple, Union
import os, json, pickle, time

# Pip
from noraise import noraise

import tldextract

from selenium.webdriver.firefox.webdriver import WebDriver

# Local
from .utils import Cookies
from ..models import Cookie

# ---------------------------------------------------------------------------------------------------------------------------------------- #



# --------------------------------------------------------- class: FirefoxCookies -------------------------------------------------------- #

class FirefoxCookies:

    # ------------------------------------------------------ Public properties ------------------------------------------------------- #

    driver             : WebDriver
    pickle_cookies     : bool
    cookies_folder_path: str


    # -------------------------------------------------------- Public methods -------------------------------------------------------- #

    def login_via_cookies(
        self,
        url: str,
        accepted_cookie_names: Union[str, List[str]]
    ) -> bool:
        org_url = self.driver.current_url
        self.driver.get(url)
        time.sleep(0.5)

        @noraise(default_return_value=False)
        def __login_via_cookies(accepted_cookie_names: Union[str, List[str]]):
            if isinstance(accepted_cookie_names, str):
                accepted_cookie_names = [accepted_cookie_names]

            for cookie in self.get_cookies():
                if accepted_cookie_names and cookie.name in accepted_cookie_names and not cookie.is_session_cookie and not cookie.is_expired:
                    return True

            if self.has_cookies_for_current_website():
                self.load_cookies()
                time.sleep(1)
                self.driver.refresh()
                time.sleep(1)

            for cookie in self.get_cookies():
                if accepted_cookie_names and cookie.name in accepted_cookie_names and not cookie.is_session_cookie and not cookie.is_expired:
                    return True

        login_res = __login_via_cookies(accepted_cookie_names)

        self.driver.get(org_url)

        return login_res

    def save_cookies(
        self,
        url: Optional[str] = None,
        folder_path: Optional[str] = None,
        cookies: Optional[List[Union[Cookie, Dict[str, any]]]] = None
    ) -> None:
        return Cookies.save(
            url=url or self.driver.current_url,
            folder_path=folder_path or self.cookies_folder_path,
            cookies=cookies or self._get_cookies(),
            pickled=self.pickle_cookies
        )

    def load_cookies(
        self,
        folder_path: Optional[str] = None,
    ) -> bool:
        if not self.has_cookies_for_current_website():
            print('Did not find any cookies for url: \'{}\''.format(self.driver.current_url))

            return False

        cookies = Cookies.load(
            url=self.driver.current_url,
            folder_path=folder_path or self.cookies_folder_path
        )

        if not cookies:
            print('Loaded cookies are None or empty for url: \'{}\''.format(self.driver.current_url))

            return False

        @noraise(default_return_value=False)
        def add_cookie(cookie: Cookie) -> bool:
            self.driver.add_cookie(cookie.dict)

            return True

        did_add_cookie = False

        for cookie in cookies:
            # if cookie.is_session_cookie:
            #     print('Cookie named \'{}\' for domain \'{}\' was a session only cookie so won\'t be used'.format(
            #         cookie.name,
            #         cookie.domain
            #     ))

            #     continue

            # if cookie.is_expired:
            #     print('Cookie named \'{}\' for domain \'{}\' did expire'.format(cookie.name, cookie.domain))

            #     continue

            if add_cookie(cookie):
                did_add_cookie = True
            else:
                print('Error while loading cookie named \'{}\' for domain \'{}\'\nURL: \'{}\'\nCookie JSON: {}'.format(
                    cookie.name,
                    cookie.domain,
                    self.driver.current_url,
                    json.dumps(cookie, indent=4)
                ))

        return did_add_cookie

    def delete_cookies(
        self,
        folder_path: Optional[str] = None,
    ) -> bool:
        return Cookies.delete(
            url=self.driver.current_url,
            folder_path=self.cookies_folder_path
        )

    def get_cookies(self) -> List[Cookie]:
        return [Cookie(d) for d in self._get_cookies()]

    def get_usable_cookies(self) -> List[Cookie]:
        return [c for c in [Cookie(d) for d in self._get_cookies()] if not c.is_expired]

    def has_all_cookies(self, cookie_names: Union[str, List[str]]) -> bool:
        existing_cookie_names = [c.name for c in self.get_usable_cookies()]

        if isinstance(cookie_names, str):
            cookie_names = [cookie_names]

        for cookie_name in cookie_names:
            if cookie_name not in existing_cookie_names:
                return False

        return True

    def has_any_cookie(self, cookie_names: Union[str, List[str]]) -> bool:
        existing_cookie_names = [c.name for c in self.get_usable_cookies()]

        if isinstance(cookie_names, str):
            cookie_names = [cookie_names]

        for cookie_name in cookie_names:
            if cookie_name in existing_cookie_names:
                return True

        return False

    def has_cookie(self, cookie_name: str) -> bool:
        return cookie_name in [c.name for c in self.get_usable_cookies()]

    def has_saved_cookies_for_current_website(self) -> bool:
        return self.has_cookies_for_url(url=self.driver.current_url)

    # alias - kept for convenience
    has_cookies_for_current_website = has_saved_cookies_for_current_website

    def has_saved_cookies_for_url(
        self,
        url: str
    ) -> bool:
        return Cookies.has_saved_cookies(
            url=url,
            folder_path=self.cookies_folder_path
        )

    # alias - kept for convenience
    has_cookies_for_url = has_saved_cookies_for_url


    # ------------------------------------------------------- Private methods -------------------------------------------------------- #

    @noraise(default_return_value=[])
    def _get_cookies(self) -> List[dict]:
        return self.driver.get_cookies()


# ---------------------------------------------------------------------------------------------------------------------------------------- #